@@ -14,6 +14,7 @@
 #include <QtCore/QPermission>
 #include <QtCore/QCoreApplication>
 #include <QtCore/QTimeZone>
+#include <QtCore/QSet>
 #include <android/log.h>
 
 Q_DECLARE_JNI_CLASS(QtPositioning, "org/qtproject/qt/android/positioning/QtPositioning")
@@ -191,6 +192,11 @@ namespace AndroidPositioning {
         QJniObject jniProvidersObj =
                 QJniObject::callStaticMethod<jobject>(positioningClass(), providerListMethodId);
         jintArray jProviders = jniProvidersObj.object<jintArray>();
+        if (!jProviders) {
+            // Work-around for QTBUG-116645
+            __android_log_print(ANDROID_LOG_INFO, logTag, "Got null providers array!");
+            return ret;
+        }
         jint *providers = env->GetIntArrayElements(jProviders, nullptr);
         const int size = env->GetArrayLength(jProviders);
         for (int i = 0; i < size; i++) {
@@ -234,6 +240,16 @@ namespace AndroidPositioning {
             if (!qFuzzyIsNull(value))
                 coordinate.setAltitude(value);
         }
+        // MSL altitude, available in API Level 34+.
+        // It will be available only if we requested it when starting updates.
+        if (QNativeInterface::QAndroidApplication::sdkVersion() >= 34) {
+            attributeExists = jniObject.callMethod<jboolean>("hasMslAltitude");
+            if (attributeExists) {
+                const jdouble value = jniObject.callMethod<jdouble>("getMslAltitudeMeters");
+                if (!qFuzzyIsNull(value))
+                    coordinate.setAltitude(value);
+            }
+        }
 
         info.setCoordinate(coordinate);
 
@@ -286,10 +302,18 @@ namespace AndroidPositioning {
         return info;
     }
 
+    using UniqueId = std::pair<int, int>;
+    static UniqueId getUid(const QGeoSatelliteInfo &info)
+    {
+        return std::make_pair(static_cast<int>(info.satelliteSystem()),
+                              info.satelliteIdentifier());
+    }
+
     QList<QGeoSatelliteInfo> satelliteInfoFromJavaLocation(JNIEnv *jniEnv,
                                                            jobjectArray satellites,
                                                            QList<QGeoSatelliteInfo>* usedInFix)
     {
+        QSet<UniqueId> uids;
         QList<QGeoSatelliteInfo> sats;
         jsize length = jniEnv->GetArrayLength(satellites);
         for (int i = 0; i<length; i++) {
@@ -340,7 +364,12 @@ namespace AndroidPositioning {
             // determining the position.
             const jboolean inFix = jniObj.callMethod<jboolean>("usedInFix");
 
+            const UniqueId id = getUid(info);
+            if (uids.contains(id))
+                continue;
+
             sats.append(info);
+            uids.insert(id);
 
             if (inFix)
                 usedInFix->append(info);
@@ -354,6 +383,7 @@ namespace AndroidPositioning {
     {
         QJniObject jniStatus(gnssStatus);
         QList<QGeoSatelliteInfo> sats;
+        QSet<UniqueId> uids;
 
         const int satellitesCount = jniStatus.callMethod<jint>("getSatelliteCount");
         for (int i = 0; i < satellitesCount; ++i) {
@@ -386,7 +416,12 @@ namespace AndroidPositioning {
             // determining the position.
             const jboolean inFix = jniStatus.callMethod<jboolean>("usedInFix", i);
 
+            const UniqueId id = getUid(info);
+            if (uids.contains(id))
+                continue;
+
             sats.append(info);
+            uids.insert(id);
 
             if (inFix)
                 usedInFix->append(info);
@@ -395,7 +430,8 @@ namespace AndroidPositioning {
         return sats;
     }
 
-    QGeoPositionInfo lastKnownPosition(bool fromSatellitePositioningMethodsOnly)
+    QGeoPositionInfo lastKnownPosition(bool fromSatellitePositioningMethodsOnly,
+                                       bool useAltitudeConverter)
     {
         QJniEnvironment env;
         if (!env.jniEnv())
@@ -405,7 +441,8 @@ namespace AndroidPositioning {
             return {};
 
         QJniObject locationObj = QJniObject::callStaticMethod<jobject>(
-                positioningClass(), lastKnownPositionMethodId, fromSatellitePositioningMethodsOnly);
+                positioningClass(), lastKnownPositionMethodId, fromSatellitePositioningMethodsOnly,
+                useAltitudeConverter);
         jobject location = locationObj.object();
         if (location == nullptr)
             return QGeoPositionInfo();
@@ -441,7 +478,7 @@ namespace AndroidPositioning {
             int errorCode = QJniObject::callStaticMethod<jint>(
                     positioningClass(), startUpdatesMethodId, androidClassKey,
                     positioningMethodToInt(source->preferredPositioningMethods()),
-                    source->updateInterval());
+                    source->updateInterval(), source->useAltitudeConverter());
             switch (errorCode) {
             case 0:
             case 1:
@@ -478,7 +515,7 @@ namespace AndroidPositioning {
             int errorCode = QJniObject::callStaticMethod<jint>(
                     positioningClass(), requestUpdateMethodId, androidClassKey,
                     positioningMethodToInt(source->preferredPositioningMethods()),
-                    timeout);
+                    timeout, source->useAltitudeConverter());
             switch (errorCode) {
             case 0:
             case 1:
@@ -550,7 +587,7 @@ static void positionUpdated(JNIEnv *env, jobject thiz, QtJniTypes::Location loca
 {
     Q_UNUSED(env);
     Q_UNUSED(thiz);
-    QGeoPositionInfo info = AndroidPositioning::positionInfoFromJavaLocation(location);
+    QGeoPositionInfo info = AndroidPositioning::positionInfoFromJavaLocation(location.object());
 
     QGeoPositionInfoSourceAndroid *source = AndroidPositioning::idToPosSource()->value(androidClassKey);
     if (!source) {
@@ -635,7 +672,7 @@ static void satelliteGnssUpdated(JNIEnv *env, jobject thiz, QtJniTypes::GnssStat
 
     QList<QGeoSatelliteInfo> inUse;
     QList<QGeoSatelliteInfo> sats =
-            AndroidPositioning::satelliteInfoFromJavaGnssStatus(gnssStatus, &inUse);
+            AndroidPositioning::satelliteInfoFromJavaGnssStatus(gnssStatus.object(), &inUse);
 
     notifySatelliteInfoUpdated(sats, inUse, androidClassKey, isSingleUpdate);
 }
@@ -675,10 +712,12 @@ static bool registerNatives()
 
     GET_AND_CHECK_STATIC_METHOD(providerListMethodId, "providerList", jintArray);
     GET_AND_CHECK_STATIC_METHOD(lastKnownPositionMethodId, "lastKnownPosition",
-                                QtJniTypes::Location, bool);
-    GET_AND_CHECK_STATIC_METHOD(startUpdatesMethodId, "startUpdates", jint, jint, jint, jint);
+                                QtJniTypes::Location, bool, bool);
+    GET_AND_CHECK_STATIC_METHOD(startUpdatesMethodId, "startUpdates", jint, jint, jint, jint,
+                                bool);
     GET_AND_CHECK_STATIC_METHOD(stopUpdatesMethodId, "stopUpdates", void, jint);
-    GET_AND_CHECK_STATIC_METHOD(requestUpdateMethodId, "requestUpdate", jint, jint, jint, jint);
+    GET_AND_CHECK_STATIC_METHOD(requestUpdateMethodId, "requestUpdate", jint, jint, jint, jint,
+                                bool);
     GET_AND_CHECK_STATIC_METHOD(startSatelliteUpdatesMethodId, "startSatelliteUpdates",
                                 jint, jint, jint, bool);
 
